﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;

namespace HeuristicLab.Algorithms.DataAnalysis.Symbolic {
  [StorableClass]
  public sealed class ConsecutiveSamplesEvaluator : SingleSuccessorOperator, IIterationBasedOperator, ISymbolicDataAnalysisIslandGeneticAlgorithmEvaluator {
    private const string ProblemDataParameterName = "ProblemData";
    private const string EvaluatorParameterName = "ProblemEvaluator";
    private const string QualityParameterName = "Quality";
    private const string FitnessCalculationPartitionParameterName = "FitnessCalculationPartition";
    private const string FixedSamplesPartitionParameterName = "FixedSamplesPartition";
    private const string ConsecutiveSamplesParameterName = "ConsecutiveSamples";
    private const string OverlapParameterName = "Overlap";
    private const string DataMigrationIntervalParameterName = "DataMigrationInterval";
    private const string IslandIndexParameterName = "IslandIndex";
    private const string IterationsParameterName = "Iterations";
    private const string MaximumIterationsParameterName = "Maximum Iterations";


    #region parameter properties
    public ILookupParameter<IDataAnalysisProblemData> ProblemDataParameter {
      get { return (ILookupParameter<IDataAnalysisProblemData>)Parameters[ProblemDataParameterName]; }
    }
    public ILookupParameter<IOperator> EvaluatorParameter {
      get { return (ILookupParameter<IOperator>)Parameters[EvaluatorParameterName]; }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters[QualityParameterName]; }
    }
    public ILookupParameter<IntRange> FitnessCalculationPartitionParameter {
      get { return (IValueLookupParameter<IntRange>)Parameters[FitnessCalculationPartitionParameterName]; }
    }
    public ILookupParameter<IntRange> FixedSamplesPartitionParameter {
      get { return (ILookupParameter<IntRange>)Parameters[FixedSamplesPartitionParameterName]; }
    }
    public IFixedValueParameter<IntValue> ConsecutiveSamplesParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[ConsecutiveSamplesParameterName]; }
    }
    public IFixedValueParameter<IntValue> OverlapParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[OverlapParameterName]; }
    }
    public IValueLookupParameter<IntValue> DataMigrationIntervalParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters[DataMigrationIntervalParameterName]; }
    }
    public ILookupParameter<IntValue> IslandIndexParameter {
      get { return (ILookupParameter<IntValue>)Parameters[IslandIndexParameterName]; }
    }
    public ILookupParameter<IntValue> IterationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters[IterationsParameterName]; }
    }
    public IValueLookupParameter<IntValue> MaximumIterationsParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters[MaximumIterationsParameterName]; }
    }

    #endregion

    #region properties

    public int ConsecutiveSamples {
      get { return ConsecutiveSamplesParameter.Value.Value; }
      set { ConsecutiveSamplesParameter.Value.Value = value; }
    }

    public int Overlap {
      get { return OverlapParameter.Value.Value; }
      set { OverlapParameter.Value.Value = value; }
    }
    #endregion

    [StorableConstructor]
    private ConsecutiveSamplesEvaluator(bool deserializing) : base(deserializing) { }
    private ConsecutiveSamplesEvaluator(ConsecutiveSamplesEvaluator original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new ConsecutiveSamplesEvaluator(this, cloner);
    }

    public ConsecutiveSamplesEvaluator()
      : base() {
      Parameters.Add(new LookupParameter<IDataAnalysisProblemData>(ProblemDataParameterName, "The problem data on which the symbolic data analysis solution should be evaluated."));
      Parameters.Add(new LookupParameter<IOperator>(EvaluatorParameterName, "The evaluator provided by the symbolic data analysis  problem."));
      Parameters.Add(new LookupParameter<DoubleValue>(QualityParameterName, "The quality which is calculated by the encapsulated evaluator."));
      Parameters.Add(new LookupParameter<IntRange>(FitnessCalculationPartitionParameterName, "The data partition used to calculate the fitness"));
      Parameters.Add(new LookupParameter<IntRange>(FixedSamplesPartitionParameterName, "The data partition which is used to calculate the fitness on the fixed samples."));
      Parameters.Add(new FixedValueParameter<IntValue>(ConsecutiveSamplesParameterName, "The number of consecutive samples used for fitness calculation in each island.", new IntValue()));
      Parameters.Add(new FixedValueParameter<IntValue>(OverlapParameterName, "The overlap for the consecutive samples used for every island.", new IntValue()));
      Parameters.Add(new ValueLookupParameter<IntValue>(DataMigrationIntervalParameterName, "The number of generations that should pass between data migration phases."));
      Parameters.Add(new LookupParameter<IntValue>(IslandIndexParameterName, "The index of the current island."));
      Parameters.Add(new LookupParameter<IntValue>(IterationsParameterName, "The number of performed iterations."));
      Parameters.Add(new ValueLookupParameter<IntValue>(MaximumIterationsParameterName, "The maximum number of performed iterations.") { Hidden = true });
    }

    public override IOperation Apply() {
      var evaluator = EvaluatorParameter.ActualValue;
      var problemData = ProblemDataParameter.ActualValue;
      var samples = FitnessCalculationPartitionParameter.ActualValue;
      var fixedSamples = FixedSamplesPartitionParameter.ActualValue;

      //create fixed rows enumerable
      var rows = Enumerable.Range(fixedSamples.Start, fixedSamples.Size);
      //create consecutive rows enumerable
      if (ConsecutiveSamples > 0) {
        var dataMigrationInterval = DataMigrationIntervalParameter.ActualValue.Value;
        var islandIndex = IslandIndexParameter.ActualValue.Value;
        var generation = IterationsParameter.ActualValue.Value;
        var iteration = islandIndex + (generation / dataMigrationInterval);
        var consecutiveRows = GenerateRows(samples, fixedSamples, ConsecutiveSamples, Overlap, iteration);
        rows = rows.Concat(consecutiveRows);
      }
      //filter out test rows
      rows = rows.Where(r => r < problemData.TestPartition.Start || r > problemData.TestPartition.End);

      //execution context is created manually to be able to clear the rows parameter easily
      var executionContext = new ExecutionContext(ExecutionContext, evaluator, ExecutionContext.Scope);

      //TODO change to lookup parameter
      executionContext.Scope.Variables.Remove("Rows");
      executionContext.Scope.Variables.Add(new HeuristicLab.Core.Variable("Rows", new EnumerableItem<int>(rows)));
      var successor = evaluator.Execute(executionContext, this.CancellationToken);
      return new OperationCollection(successor, base.Apply());
    }

    public static IEnumerable<int> GenerateRows(IntRange samples, IntRange fixedSamples, int consecutiveSamples, int overlap, int iteration) {
      var consecutiveSamplesStart = (consecutiveSamples - overlap) * iteration;
      consecutiveSamplesStart = consecutiveSamplesStart % (samples.Size - fixedSamples.Size);
      var rows = Enumerable.Range(fixedSamples.End, samples.End - fixedSamples.End);
      rows = rows.Concat(Enumerable.Range(samples.Start, fixedSamples.Start - samples.Start));
      rows = rows.Concat(Enumerable.Range(fixedSamples.End, samples.End - fixedSamples.End));
      rows = rows.Concat(Enumerable.Range(samples.Start, fixedSamples.Start - samples.Start));
      return rows.Skip(consecutiveSamplesStart).Take(consecutiveSamples);
    }
  }
}
